from argparse import ArgumentParser
from core import BSPInterface, Backup
from bsplzma import BSPLZMA
from delete_l27 import DeleteL27
from config import Config
from pathlib import Path
from os import path
from time import time

argparser = ArgumentParser(description="Parse or pack a BSP file",
                                        epilog="-file, -dir, and -noauto will have no effect unless -pack is present",
                                        usage="%(prog)s [-h] path [-parse] [-pack] [-noauto] [-file FILE] [-dir DIRECTORY]")

argparser.add_argument("path", type=str,
                                   help="The path of the target BSP.")
argparser.add_argument("-parse", dest="parse", default=False, action="store_true",
                                   help="Parse the BSP and display information on files that would need to be packed, and files currently packed.")
argparser.add_argument("-pack", dest="pack", default=False, action="store_true",
                                   help="Pack files or directories into the BSP.")
##argparser.add_argument("-repack", dest="repack", default=False, action="store_true",
##                                   help="Repack the BSP")
## not currently implemented fully ##
argparser.add_argument("-rl27", dest="rl27", default=False, action="store_true",
                                   help="Remove bsp lump 27")                  


args = argparser.parse_args()
config = Config(Path(__file__).parent)

bsp_path = args.path.strip("/\\")
    
try:
    bsp_path.relative_to(config.getkv("tf_path"))
except:
    bsp_path = Path(config.getkv("tf_path")) / bsp_path
if not(path.exists(bsp_path)):
    raise IOError("BSP %s not found" %bsp_path)


with open(bsp_path, "r+b") as fp:
    starttime = time()

    location = Path(__file__).parent
    
    fdlist = [location / "fdlists/all.fdl", location / "fdlists" / Path(bsp_path).with_suffix(".fdl").name]
    interface = BSPInterface(fp, fdlist, location)
    
    if args.parse:
        interface.list_needed_assets()
        parsetime = time()
        print("Parse finished in %s seconds" %round(parsetime-starttime, 2))
    else:
        parsetime = time()

    try:
        if args.pack or args.rl27: # or args.repack:
            Backup.backup_file(fp.name)

        if args.rl27:
            remover = DeleteL27(fp)
            remover.delete_l27()

        if args.pack:
            interface.autopack_assets()
                
            packtime = time()
            print("Packing finished in %s seconds" %round(packtime-parsetime, 2))

##        if args.repack:
##            lzma = BSPLZMA(fp)
##            lzma.compress()
                
    except Exception as error:
        Backup.restore_backup(fp.name)
        print("\nA fatal error occured during packing. Backups restored\n")
        raise (error) from None
        
