import os
from pathlib import Path

class Author():
    def __init__(self, location, tf_path):
        self.__default_authors = location / "authors.txt"
        self.__TF_PATH = tf_path
        if not(os.path.exists(self.__default_authors)):
            afp = open(self.__default_authors, "wt")
            afp.close()

    def __return_author(self, asset, line, location):
        ##line is author name + name/dir to compare
        if not line:
            return
        index = line.find("--")
        aname = line[:index].strip(" ")
        author = "| author: " + line[index+2:].strip(" \n")


        if "*" not in aname and (self.__TF_PATH / location / aname).is_dir():
            if Path(os.path.relpath(asset, self.__TF_PATH / location / aname)).parts[0] != "..":
                ## is this asset relative to the path (in the folder)?
                return author
        
        if "*" in aname:
            if asset.match(str(location / aname)):
                return author
            else:
                parent_no = 0
                while str(location.parents[parent_no]) != ".":
                    if asset.match(str(location.parents[parent_no] / aname)):
                        return author
                    parent_no += 1
        
        if location / aname == asset:
            return author
    
    def get_author(self, asset):
        r'''pass in an asset path and either "" or the author name will be returned
        '''
        ## in an authors file, each line is set up as:
        ## <asset name> -- <author name>
        ## where any amount of whitespaces can be around any asset name
        ## alternatively, you may define wildcards as such:
        ## * -- <author name> for all assets,
        ## or <partial asset name>*<partial asset name> -- <author name> for partial wildcarding of assets
        if not asset:
            return ""

        apath = asset.relative_to(self.__TF_PATH)
        parent_no = 0
        ##while not located in the /tf/ folder
        while str(apath.parents[parent_no]) != ".":
            location = apath.parents[parent_no]
            
            if os.path.exists(self.__TF_PATH / location / "authors.txt"):
                with open(self.__TF_PATH / location / "authors.txt", "rt") as afp:
                    for line in afp:
                        item = self.__return_author(asset, line, location)
                        if item: return item
            parent_no += 1

        ## default file does not take priority over the in-folder one
        with open(self.__default_authors, "rt") as afp:
            for line in afp:
                item = self.__return_author(asset, line, asset.parent.relative_to(self.__TF_PATH))
                if item: return item
        
        return ""

                        
                        
