import os
from .file_family import FileFamilyManager
from .mdlfile import MDLparser
from .pcffile import PCFparser
from pathlib import Path

method_list = [x for x in dir(Path) if callable(getattr(Path, x)) and not x.startswith("__")]

if "is_relative_to" not in method_list:
    print("Python 3.7 detected. Amending methods")
    def is_relative_to(self, *other):
        try:
            self.relative_to(*other)
            return True
        except ValueError:
            return False
    Path.is_relative_to = is_relative_to

    


class FormatAssets():
    def __init__(self, asset_paths, tf_path, bsp_fp, remove_assets, *args):
        print("formatting assets..")
        r'''pass in an array of relative path objects and the /tf/ folder absolute path
        '''
        self.__TF_PATH = tf_path
        self.__bsp_fp = bsp_fp
        self.__skin_dicts = args
        paths_arr, fixed_paths = [], []

        for path in asset_paths:
            paths_arr.append(self.__TF_PATH / path)
            
        if os.name == "posix":
            divider = "/"
        else:
            divider = "\\"

        self.__pcf_index = {}
        self.__used_pcfs = []
        ##build list of particles
        if os.path.exists(self.__TF_PATH / "particles"): ##does the particles folder actually exist?
            for file in (item for item in (self.__TF_PATH / "particles").iterdir() if item.is_file()):
                self.__parse_particles(file)
                
        for folder in (item for item in (self.__TF_PATH / "custom").iterdir() if item.is_dir()):
            if os.path.exists(folder / "particles"):
                for file in (folder / "particles").iterdir():
                    self.__parse_particles(file)
         
        
        for path in paths_arr:
            ## - team materials - ##
            if "*" in str(path):
                paths = path.parent.glob(path.name)
                if paths:
                    for item in paths:
                        fixed_paths.append(item) ##test me
                else:
                    print("Does this wildcard not work? Test for %s" %str(path))
            ## - res files - ##
            elif path.suffix == ".res": ##find someone to do testing with
                print(".res file found - contact me to help test packing with this format (minimal work needed)")
                if os.path.exists(path):
                    with open(path, "rt") as rfp:
                        for line in rfp:
                            if '"image"' in line:
                                path = line.replace(" ", "")[7:]
                                path = (self.__TF_PATH / path.replace("..", "")) ##test me
            ## - sounds - ##
            elif path.suffix in [".mp3", ".wav"]:
                fname = path.relative_to(self.__TF_PATH)
                fixed_paths.append(self.__TF_PATH / str(fname).replace(" ", "_"))       
            
            ## - skies - ##
            elif path.suffix == ".is_sky":
                parent = path.parent
                end = path.stem
                for side in ["bk", "dn", "ft", "lf", "rt", "up"]:
                    fixed_paths.append(parent / "materials" / "skybox" / (end+side+".vmt"))
                    fixed_paths.append(parent / "materials" / "skybox" / (end+"_hdr"+side+".vmt"))
                    #hdr may not be in the BSP but skies included anyway for now
            elif path.suffix == ".is_detailsprite":
                fixed_paths.append(self.__TF_PATH / "materials" / "detail" / (path.stem+".vmt"))
            ## - particles - ##
            elif path.suffix == ".in_pcf":
                for entry in self.__pcf_index:
                    if path.stem in self.__pcf_index[entry]:
                        fixed_paths.append(self.__TF_PATH / entry)
                        self.__used_pcfs.append(entry)
            ## - sound scripts and scapes- ##
            elif path.suffix == ".in_sound_script":
                for file in (item for item in (self.__TF_PATH / "scripts").iterdir() if item.is_file()):
                    with open(file, "rt") as sfp:
                        data = sfp.read()
                    if path.stem in data:
                        fixed_paths.append(self.__TF_PATH / "scripts" / file.name) ##test me
            ## - quick fix for colour correction - ##
            elif path.suffix == ".raw":
                if Path(path.stem).suffix == ".raw":
                    fixed_paths.append(path.parent / path.stem)
                else:
                    fixed_paths.append(path)
            ## - team icons may have no ending or .vtf - 
            elif path.suffix == ".team_icon":
                if path.suffixes[0] == ".vtf":
                    fixed_paths.append(path.parent / path.stem)
                else:
                    fixed_paths.append(path.parent / (path.stem + ".vtf"))
                
            
            ## - cubemaps/wvt patches - ##
            elif path.is_relative_to(self.__TF_PATH / "materials" / "maps"):
                asset = path.name
                parent = path.parents[3] / path.parent.relative_to(path.parents[1])
                if "_wvt_patch" in asset:
                    fixed_paths.append(parent / asset.replace("_wvt_patch", ""))
                else:
                    index = asset.rfind("_", 0, asset.rfind("_", 0, asset.rfind("_"))) #format is matname_*_*_* so this removes unwanted info
                    index2 = asset.find(".") ##start of extension
                    if index == -1:
                        raise ValueError("Issue with cleaning cubemap materials")
                    fixed_paths.append(parent / (asset[:index] + asset[index2:]))

            ## - rope material - ##
            ##rope materials may have .vmt or not - testing for this
            elif path.suffix == ".rope_material":
                if path.suffixes[0] == ".vmt": ## if it has .vmt, format is name.vmt.rope_material and we test for this here easily
                    path = path.parent / path.stem
                else:
                    path = path.parent / (path.stem + ".vmt")
                fixed_paths.append(path)
                    
            elif path.suffix == "":
                raise ValueError("Path %s has no suffix" %path.relative_to(self.__TF_PATH))
            elif len(path.suffixes) > 1:
                raise ValueError("Path %s has too many suffixes" %path.relative_to(self.__TF_PATH))
            else:
                fixed_paths.append(path)

##final fixups section##

        ##add mvm population files and other map scripts
        mapname = str(Path(self.__bsp_fp.name).stem)
        for item in Path(self.__bsp_fp.name).parent.glob(mapname + "_*"):
            if item.suffix != ".bsp":
                fixed_paths.append(item)
        path = self.__TF_PATH / "materials" / "vgui" / "maps" / ("menu_photos_" + mapname + ".vmt")
        fixed_paths.append(path)
        ##
                
        ##final pcf fixup
        self.__used_pcfs = set(self.__used_pcfs)
        if self.__used_pcfs:
            manifest_name = self.__generate_manifest(self.__used_pcfs,
                                Path(self.__bsp_fp.name).stem)
            fixed_paths.append(manifest_name)
        ##

        ##remove items specified in fdlists
        fixed_paths_ra = []
        for path in fixed_paths:
            for item in remove_assets:
                if path.match(str(self.__TF_PATH / item)):
                    break
            else:
                fixed_paths_ra.append(path)
        fixed_paths = fixed_paths_ra
        ##
        ##custom files from soundscapes
        for item in set(fixed_paths):
            if os.path.exists(item) and item.relative_to(self.__TF_PATH).parent == Path("scripts"):
                with open(item, "rt") as sfp:
                    for line in sfp:
                        if '"wave"' in line.replace(" ", ""):
                            sound = self.__TF_PATH / "sound" / line.replace(" ", "").replace('"wave"', "").strip("\t\n").replace('"', "")
                            ##this lengthy statement strips \t\t\t"wave" "<sound>" down to <sound>
                            fixed_paths.append(sound)



        self.__asset_families = FileFamilyManager(fixed_paths, self.__TF_PATH)

        if os.name == "posix": ##linux fixup - all files are stored as lowercase, yet on disc may be uppercase
            unique_paths = set()
            for family in self.__asset_families.file_families: #standard paths
                if os.path.exists(self.__TF_PATH / family.path):
                    for asset in family.assets:
                        unique_paths.add(self.__TF_PATH / family.path / Path(asset).parent)
            found_assets = []

            for path in unique_paths:
                for asset in (asset for asset in path.iterdir() if asset.is_file()):
                    if (str(asset.name).lower()) != str(asset.name) and \
				self.__asset_families.search_asset(asset.parent / str(asset.name).lower()): ##lowercase fname in this asset family?
                        print("uppercase asset name found, fixing")
                        self.__asset_families.remove_asset(asset.parent / str(asset.name).lower())
                        self.__asset_families.add_asset(asset) ##replacing with the uppercase one

        self.__unique_assets = []
        for family in self.__asset_families.file_families: #standard paths
            if os.path.exists(self.__TF_PATH / family.path):
                for asset in family.assets:
                    if os.path.exists(self.__TF_PATH / family.path / asset):
                        self.__unique_assets.append(self.__TF_PATH / family.path / asset)
                        
        for folder in (folder for folder in (self.__TF_PATH / "custom").iterdir() if folder.is_dir()):
            for family in self.__asset_families.file_families:
            ##all subdirectories in /tf/custom
                for asset in family.assets:
                    if os.path.exists(folder / family.path / asset):
                        self.__unique_assets.append(folder / family.path / asset)

        print("done")

        ##fixup VMTs and MDLs properly
        output_assets = []
        search_assets = [(b"$fallbackmaterial", ".vmt"), (b"$bumpmap", ".vtf"),
                         (b"$normalmap", ".vtf"), (b"$bottommaterial", ""),
                         (b"%tooltexture", ".vtf"), (b"$basetexture", ".vtf"),
                         (b"$selfillummask", ".vtf"), (b"$texture2", ".vtf"),
                         (b"$basetexture2", ".vtf"), (b"$bumpmap2", ".vtf")]
        max_read_len = 96
        offset = 0

        print("collecting references..")
        
        ## collect referenced/alike assets here
        for path in self.__unique_assets:
            ## - models - ##
            if path.suffix == ".mdl":
                if os.name == "posix":
                    for asset in (asset for asset in path.parent.iterdir() if \
                                              asset.is_file() and \
                                              str(path.stem) == str(Path(asset.stem).stem).lower()):
                        if asset.suffix != ".mdl":
                            output_assets.append(asset)
                else:
                    for item in [".vvd", ".sw.vtx", ".dx80.vtx", ".dx90.vtx"]:
                        output_assets.append(path.parent / (path.stem + item))
                    for item in [".phy", ".xbox.vtx"]:
                        if os.path.exists(path.parent / (path.stem + item)):
                            output_assets.append(path.parent / (path.stem + item)) ##these two may not exist

                mdlfile = MDLparser(self.__TF_PATH, self.__TF_PATH / path)
                if Path(os.path.relpath(path, self.__TF_PATH / "custom")).parts[0] != "..":
                    ## the above statement figures out if it is a /custom/ folder or not
                    ## as if /tf/custom/ is at the end, the final path component is not ..
                    custompath = self.__TF_PATH / "custom" / (path.relative_to(self.__TF_PATH).parts[1])
                    materials = [(custompath / apath) for apath in mdlfile.get_materials(self.__skin_dicts).copy()]
                else:
                    materials = map(lambda path : self.__TF_PATH / path, mdlfile.get_materials(self.__skin_dicts).copy())
                if os.name == "posix":
              	     for item in materials:
              	         self.__unique_assets.append(Path(str(item).replace("\\", "/")))
                else:
              	     self.__unique_assets += materials
                del mdlfile
                
        for fpath in self.__unique_assets:
            ## two separate loops so we can add materials from models
            ## - vmt references - ##
            if fpath.suffix == ".vmt" and os.path.exists(fpath):
                with open(fpath, "rb") as fp:
                    ##simplified version of lump 0 parsing
                    data = fp.read()
                    for item in search_assets:
                        offset = data.lower().find(item[0])
                        if offset != -1:
                            fp.seek(offset)
                            string = fp.read(max_read_len).decode().split("\n")[0].encode()
                            if string != string.replace(b"\t", b""):
                                string = string.replace(b"\t", b"")
                                string = string.replace(b'""', b'" "')
                                
                            if data.lower().find(b"\"%s\"" %item[0]) == -1:
                                index1 = len(item[0]) + 2 
                            else:
                                index1 = len(item[0]) + 3
                                ## sometimes the tag has "" around it, sometimes not
                                ## todo - look at different ways people might format VMTs
                                ## to allow for parsing materials from them even if weirdly formatted

                            index2 = string.find(b'"', index1)
                            
                            if Path(os.path.relpath(fpath, self.__TF_PATH / "custom")).parts[0] != "..":
                                #if /custom/
                                custompath = self.__TF_PATH / "custom" / (fpath.relative_to(self.__TF_PATH).parts[1])
                                asset = (custompath / Path("materials") / (string[index1:index2].decode() + item[1]))

                            else:
                                asset = (self.__TF_PATH / Path("materials") /
                                        (string[index1:index2].decode() + item[1]))
                                if os.name == "posix":
                                    asset = Path(str(asset).replace("\\", "/")) ##vmt paths may use \ instead of / on linux
                                
                            if os.path.exists(asset):
                                output_assets.append(asset)
                            else:
                                asset = self.__TF_PATH / Path(str(asset.relative_to(self.__TF_PATH)).lower())
                                if os.path.exists(asset):
                                    output_assets.append(asset)
                            



            if os.path.exists(fpath):
                output_assets.append(fpath)
            self.__unique_assets = output_assets
        print("done")
    
    def __generate_manifest(self, particles, mapname):
        ##takes in an array of particles
        mapname = mapname + "_particles.txt"
        if not os.path.exists(self.__TF_PATH / "maps" / mapname):
            with open(self.__TF_PATH / "maps" / mapname, "wt") as mfp:
                mfp.write("particles_manifest\n{\n")
                for item in particles:
                    mfp.write("\tfile\t\"particles/%s.pcf\"\n" %item.stem)
                mfp.write("}")
        return (self.__TF_PATH / "maps" / mapname)

    def __parse_particles(self, file):
        if file.suffix == ".pcf":
            parser = PCFparser(file)
            self.__pcf_index[file.relative_to(self.__TF_PATH)] = parser.get_particles()

    def get_needed_assets(self):
        return self.__unique_assets

    
