from .structure_parser import Struct
from .lump_class_base import LumpBaseParser
import re
import time

class Lump0Parser(LumpBaseParser):
    def __init__(self, fp, length):
        super().__init__(fp, length)

        self.__BUFFER_SIZE = 4096
        self.__LOOKBACK_WINDOW_SIZE = 48 #looks back so that strings are captured if split at the end of the buffer
        self.__MAX_READ_LEN = 160 #how many bytes to read looking for each item's material/etc, larger values slow the search but potentially catch longer names

        if self._lump_len < (self.__BUFFER_SIZE-self.__LOOKBACK_WINDOW_SIZE):
            self.__BUFFER_SIZE = self._lump_len

    
        #based on the default fgd for tf2
        self.__search_once = [b'"skyname"', b'"detailmaterial"'] #all string searched for only once
        self.__search_repeats = [b'"material"', b'"model"\x20"models', b'"overlaymaterial"',
                                 b'"RopeMaterial"', b'"team_model"', b'"team_icon"', b'"PlayVO"',
                                 b'"flag_icon"', b'"flag_paper"', b'"flag_trail"', b'"powerup_model"',
                                 b'"pickup_model"', b'"pickup_particle"', b'"point_warn_sound"',
                                 b'"icon"', b'"hud_icon"', b'"res_file"', b'"hud_res_file"',
                                 b'"ParticleEffect"', b'"gibmodel"', b'"filename"', b'"effect_name"',
                                 b'"team_model_0"', b'"team_model_2"', b'"team_model_3"',
                                 b'"team_icon_0"', b'"team_icon_2"', b'"team_icon_3"',
                                 b'"team_overlay_0"', b'"team_overlay_2"', b'"team_overlay_3"',
                                 b'"message"', b'"soundscape"'] #all strings searched multiple times for
        ##add team_capsound_* later, once extension is found
        ##extensions stored separately for mapping
        self.__so_extensions = [b".IS_SKY", b".IS_DETAILSPRITE"]
        ##detail vbsp likely does not have to be packed. I have not included it but this may need to change
        self.__sr_extensions = [b".vmt", None, b".vmt", 
                                b".ROPE_MATERIAL", None, b".vmt", None,
                                b"_*.vmt", b".IN_PCF", b"_*.vmt", None,
                                None, None, b".IN_SOUND_SCRIPT",
                                b".vmt", None, None, None,
                                b".IN_PCF", None, b".raw", b".IN_PCF",
                                None, None, None,
                                b".TEAM_ICON", b".TEAM_ICON", b".TEAM_ICON",
                                b".TEAM_ICON", b".TEAM_ICON", b".TEAM_ICON",
                                None, b".IN_SOUND_SCRIPT"]
        
        
    def __init_parse_fp(self):
        self._fp.seek(-self.__LOOKBACK_WINDOW_SIZE, 1)

    def __fixup_array(self, string):
        index = string.find(b'"\x20"') + 3 #splits each item in the lump
        index2 = string.find(b'"', index)
        if string[:9] == b'"message"':
            if b".mp3" not in string and b".wav" not in string:
                return None
        if index2 == -1:
            raise IndexError("issue parsing lump 0; increase max read len in config. Issue was while parsing %s" %string)
        return string[index:index2]

    def __fixup_array_ints(self, string):
        fixup = self.__fixup_array(string)
        if fixup:
            return int(fixup.decode())
        else: return None

    def parse_skins(self):
        print("parsing skins..")
        ents = [b'"classname"\x20"prop_dynamic"']
        kv = b"skin"
        model = b'"model" "models'
        modelarr = []
        skinarr = []
        earr = [] ##for storing the names of erroring props (errors may be normal)
        offset = 0
        max_dist_rb = 768 ##don't search further back than this for {
        max_dist_rf = 64 ##classname seems to be at the end of the entity definition, so we read forwards only a bit
        
        self._init_fp()
        data = self._fp.read(self._lump_len)

        for ent in ents:
            while offset != -1:
                offset = data.find(ent, offset+len(ent))
                if offset != -1:
                    ##find skin and model kvs
                    classdata = data[offset-max_dist_rb : offset+max_dist_rf]
                    i1, i2 = classdata.find(b"{"), classdata.rfind(b"}")
                    classdata = classdata[i1:i2]
                    mindex, sindex = classdata.find(model), classdata.find(kv)
                    
                    if mindex != -1 and sindex != -1:
                        modelarr.append(classdata[mindex:mindex+self.__MAX_READ_LEN])
                        skinarr.append(classdata[sindex:sindex+self.__MAX_READ_LEN])
                    else:
                        tindex = classdata.find(b'"targetname" "')
                        tname = classdata[tindex+14:]
                        tnindex = tname.find(b'"')
                        tname = tname[:tnindex]
                        earr.append(tname.decode())
                        #print("Issue while parsing skins of entity '%s'. Check read dist values if a skin kv is present" %tname.decode())
        print("Entities for which skins could not be found. Check to see if any known props with skins are here.")
        print("Coordinates indicate an unnamed prop_dynamic:\n %s" %earr)

        ##return a dictionary mapping the model name to a set of skins used
        output = {}
        for item in zip(map(self.__fixup_array,
                             modelarr),
                         map(lambda string : int(self.__fixup_array(string).decode()),
                             skinarr)):
            if item[0] not in output and item[0]:
                output[item[0]] = set()
            output[item[0]].add(item[1])
        print("done")
        return output
        


    def parse(self):    
        self._init_fp()
        ##search for each keyword item##
        index_dict, extension_list = [], []
        item_offset = 0
        ADJUSTED_BUFFER_SIZE = self.__BUFFER_SIZE-self.__LOOKBACK_WINDOW_SIZE #accounts for the lookback window used to ensure no missed matches
        buffer_read_amount = self._lump_len//ADJUSTED_BUFFER_SIZE #how many buffer reads to loop through
        if self._lump_len%self.__BUFFER_SIZE:
            buffer_read_amount += 1 #+1 read if the lump cannot be evenly divided


        for buffer_read in range(buffer_read_amount):
            file_data = self._fp.read(self.__BUFFER_SIZE)
            self.__init_parse_fp()
            
            for item in self.__search_once:
                extension_index = self.__search_once.index(item)
                item_offset = file_data.find(item)
                if item_offset != -1: #if it appears
                    index_dict.append(item_offset + buffer_read*(ADJUSTED_BUFFER_SIZE))
                    extension_list.append(self.__so_extensions[extension_index])
                    
            for item in self.__search_repeats:
                extension_index = self.__search_repeats.index(item)
                item_offset = 0
                while item_offset != -1:
                    item_offset = file_data.find(item, item_offset+len(item))
                    if item_offset != -1:
                        index_dict.append(item_offset+buffer_read*(ADJUSTED_BUFFER_SIZE))
                        extension_list.append(self.__sr_extensions[extension_index])

        items_array = []
        for x in index_dict:
            self._init_fp()
            self._fp.seek(self._lump_offset + x)
            items_array.append(self._fp.read(self.__MAX_READ_LEN))

        output = map(self.__fixup_array, items_array)
        output = set(map((lambda string, extension : (string+extension) if extension and string else string),
                         output, extension_list))
        output = [item for item in output if item]

        return output
