from .lumps.structure_parser import Struct
from pathlib import Path


class PCFparser():
    def __init__(self, filepath):
        self.__PCFheader = Struct("pcf_header",
                            (("id", "<!-- dmx encoding binary 2 format pcf 1 -->"),
                             ("filler", "SHORTUL"),
                             ("nStrings", "SHORTUL")))

        self.__CDmxHeader = Struct("CDmxHeader",
                            (("nEntries", "INTUL"),))
        self.__CDmxElementHeader = Struct("CDmxElementHeader",
                            (("typeNameIndex", "SHORTUL"),))
        self.__CDmxAttribute = Struct("CDmxAttribute",
                            (("typeNameIndex", "SHORTUL"),
                             ("attributeType", "CHARL"),
                             ("attributeData", "INTUL")))
                

        self.__filepath = filepath
        self.__strings_dict = [""]

    def get_particles(self):
        with open(self.__filepath, "rb") as fp:
            header = self.__PCFheader.parse(fp)

            while len(self.__strings_dict) <= header["nStrings"]:
                data = fp.read(1)
                if data != b"\x00":
                    self.__strings_dict[-1] += data.decode()
                else:
                    self.__strings_dict.append("")
            self.__strings_dict.remove("")

            ##annoyingly, the names of particles are null terminated strings.
            ##Thus, we do the following instead of parsing in a standard manner
            self.__element_info = [[0, ""],]
            for item in range(self.__CDmxHeader.parse(fp)["nEntries"]):
                self.__element_info[-1][0] = self.__CDmxElementHeader.parse(fp)["typeNameIndex"]
                ##made a struct for proper decoding
                end_found = False
                while not end_found:
                    data = fp.read(1)
                    if data != b"\x00":
                        self.__element_info[-1][1] += data.decode()
                    else:
                        self.__element_info.append([0, ""], )
                        end_found = True
                fp.read(16) ##16 charls, dataSignature

            self.__element_info.remove([0, ""],)
            output = [""]
            definition_index = self.__strings_dict.index("DmeParticleSystemDefinition")
            for item in self.__element_info:
                if item[0] == definition_index:
                    output.append(item[1])
            return output            

